from typing import Any, cast

from pydantic import BaseModel


# https://docs.pact.io/implementation_guides/jvm/matching#matcher-selection-algorithm
def generate_matching_rules(model: type[BaseModel]) -> dict[str, dict[str, str]]:
    schema = model.model_json_schema()
    prop = cast(dict[str, Any], schema.get("properties", {}))
    result: dict[str, dict[str, str]] = {}
    for field, field_schema_raw in prop.items():  # pyright: ignore[reportAny]
        field_schema = cast(dict[str, Any], field_schema_raw)
        result[f"$.body.{field}"] = {
            "match": cast(str, field_schema.get("type", "string"))
        }
    return result



from dataclasses import dataclass, field
from typing import Any, Protocol

from pydantic import BaseModel


@dataclass
class EndpointMetadata:
    http_method: str
    path: str
    return_type: type[BaseModel]


class EndpointFunction(Protocol):
    _accord_endpoint: EndpointMetadata

    def __call__(self, *args, **kwargs): ...


@dataclass
class ParsedInteraction:
    description: str = ""
    request: dict[str, str] = field(default_factory=dict)
    response: dict[str, Any] = field(default_factory=dict)
    matching_rules: dict[str, dict[str, str]] = field(default_factory=dict)


@dataclass
class ParsedContract:
    consumer: str = ""
    producer: str = ""
    interactions: list[ParsedInteraction] = field(default_factory=list)



import json
import re
import threading

from pydantic import BaseModel
from werkzeug.routing import Map, Rule
from werkzeug.serving import BaseWSGIServer, make_server
from werkzeug.wrappers import Response

from core.contract import Contract
from serialization.writer import write_contract


class ExampleBuilder:
    def __init__(self, examples: dict[str, BaseModel], name: str):
        self.examples = examples
        self.name = name

    def example(self, model: BaseModel):
        self.examples[self.name] = model


class MockServer:
    def __init__(self, contract: type[Contract]):
        self.contract: type[Contract] = contract
        self.examples: dict[str, BaseModel] = {}

        self.server: BaseWSGIServer
        self.thread: threading.Thread
        self.port: int = 5000
        self.map: Map

    def given(self, name: str) -> ExampleBuilder:
        return ExampleBuilder(self.examples, name)

    def _wsgi_app(self, environ, start_response):
        adapter = self.map.bind_to_environ(environ)
        endpoint, kwargs = adapter.match()  # pyright: ignore[reportAny]
        example = self.examples[endpoint]
        response = Response(
            json.dumps(example.model_dump()), mimetype="application/json"
        )

        return response(environ, start_response)

    def __enter__(self):
        rules = []

        for name, interaction in self.contract.interactions.items():
            path_ = re.sub(r"{(.*)}", r"<string:\1>", interaction.path)
            rules.append(Rule(path_, endpoint=name, methods=[interaction.http_method]))

        self.map = Map(rules)

        self.server = make_server("127.0.0.1", self.port, self._wsgi_app, threaded=True)
        self.thread = threading.Thread(target=self.server.serve_forever)
        self.thread.start()

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        write_contract(self.contract, self.examples)

        self.server.shutdown()
        self.thread.join()



from functools import partial, wraps
from inspect import getmembers, isfunction
from typing import Any, Callable, cast, get_type_hints

from pydantic import BaseModel

from .models import EndpointFunction, EndpointMetadata


class Contract:
    consumer: str = ""
    producer: str = ""

    def __init_subclass__(cls) -> None:
        cls.interactions: dict[str, EndpointMetadata] = {}
        class_methods = getmembers(cls, predicate=isfunction)

        for name, method in class_methods:
            if hasattr(method, "_accord_endpoint"):
                obj_temp = cast(object, method)
                endpoint_fn = cast(EndpointFunction, obj_temp)

                metadata = endpoint_fn._accord_endpoint

                cls.interactions[name] = metadata


# https://stackoverflow.com/a/10300995/9693927
def endpoint[F: Callable[..., Any]](
    func: Callable[..., Any] | None = None,
    /,
    *,
    http_method: str | None = None,
    path: str | None = None,
):
    if not func:
        return partial(endpoint, http_method=http_method, path=path)  # pyright: ignore[reportUnknownVariableType]

    if not http_method or not path:
        raise ValueError("endpoint decorator requires http_method and path")

    @wraps(func)
    def wrapper(*args, **kwargs) -> F:
        return func(*args, **kwargs)  # pyright: ignore[reportAny]

    ret = get_type_hints(func).get("return")
    if ret is None:
        raise ValueError(f"Endpoint {func.__name__} must have a return type annotation")
    ret = cast(type[BaseModel], ret)

    obj_temp = cast(object, wrapper)
    endpoint_fn = cast(EndpointFunction, obj_temp)

    endpoint_fn._accord_endpoint = EndpointMetadata(http_method, path, ret)
    return endpoint_fn



import json
from pathlib import Path
from typing import Any, cast

from core.models import ParsedContract, ParsedInteraction


def read_contract(contract_file: Path) -> ParsedContract:
    _recovered_contract = ParsedContract()

    with open(contract_file) as file:
        json_object = cast(dict[str, Any], json.loads(file.read()))

        _recovered_contract.consumer = json_object["consumer"]["name"]
        _recovered_contract.producer = json_object["provider"]["name"]

        interactions = cast(list[dict[str, Any]], json_object["interactions"])
        for interaction in interactions:
            _recovered_interaction = ParsedInteraction()

            _recovered_interaction.description = interaction["description"]
            _recovered_interaction.matching_rules = interaction["response"][
                "matchingRules"
            ]
            _recovered_interaction.request = interaction["request"]
            _recovered_interaction.response = interaction["response"]

            _recovered_contract.interactions.append(_recovered_interaction)

    return _recovered_contract



import json
from pathlib import Path
from typing import Any

from pydantic import BaseModel

from core.contract import Contract
from core.schema import generate_matching_rules


def write_contract(
    contract: type[Contract], examples: dict[str, BaseModel], output="./accords"
):
    consumer = contract.consumer
    producer = contract.producer

    interactions = []

    for name, metadata in contract.interactions.items():
        example = examples[name]
        interaction = {
            "description": metadata.path,
            "request": {"method": metadata.http_method, "path": metadata.path},
            "response": {
                "status": 200,
                "body": example.model_dump(),
                "matchingRules": generate_matching_rules(metadata.return_type),
            },
        }
        interactions.append(interaction)

    accord: dict[str, dict[str, str] | list[dict[str, Any]]] = {
        "consumer": {"name": consumer},
        "provider": {"name": producer},
        "interactions": interactions,
    }

    out = Path(output)
    out.mkdir(exist_ok=True)
    with open(out / f"{consumer}-{producer}.json", "w") as f:
        json.dump(accord, f, indent=2)



import re
from pathlib import Path
from typing import Any, cast

import httpx

from serialization.reader import read_contract

JSON_TYPE_MAP = {
    "string": str,
    "integer": int,
    "number": (int, float),
    "boolean": bool,
    "array": list,
    "object": dict,
    "null": type(None),
}


class ContractVerifier:
    def __init__(self, contract_path: Path, base_url: str) -> None:
        self.contract_path = contract_path
        self.base_url = base_url

    def verify(self) -> None:
        pc = read_contract(self.contract_path)

        for interaction in pc.interactions:
            path = interaction.request["path"]
            example_body = cast(dict[str, Any], interaction.response.get("body", {}))

            for param_name, param_value_raw in example_body.items():  # pyright: ignore[reportAny]
                param_value = cast(str | int | float, param_value_raw)
                path = re.sub(rf"{{{param_name}}}", str(param_value), path)

            res = httpx.request(
                interaction.request["method"],
                self.base_url + path,
            )

            body: dict[str, Any] = cast(dict[str, Any], res.json())
            for path, rule in interaction.matching_rules.items():
                field = path.split(".")[-1]

                if rule["match"] not in JSON_TYPE_MAP:
                    raise ValueError(f"Unsupported match type '{rule['match']}'")
                expected_type = JSON_TYPE_MAP[rule["match"]]

                actual_value = cast(object, body[field])
                if not isinstance(actual_value, expected_type):
                    raise ValueError(
                        f"Field '{field}' expected type {expected_type}, got {type(actual_value)}"
                    )
